/*
 * Define all nodes used by the interpreter.
 * It is extremly important that these nodes are as small as possible,
 * because they are the outstandingly most allocated data structure.
 *
 * Common for all nodes are:
 * type:	The type of the node.
 * line:	The line in the source. The 4 top bits store the
 *		basic type of the node (L_ something). Use L_MASK to mask
 *		out either the line number or basic type.
 */

#define L_SINGLE		0x10000
struct lnode_single {
    int line;
    struct value *(*fun)();
};

#define L_NUMBER		0x20000
struct lnode_number {
    int line;
    struct value *(*fun)();
    int number;
    double n_float;
};

#define L_NAME			0x30000
struct lnode_name {
    int line;
    struct value *(*fun)();
    char *name;
};

#define L_1			0x50000
struct lnode_1 {
    int line;
    struct value *(*fun)();
    struct lnode *expr;
};

#define L_2			0x60000
struct lnode_2 {
    int line;
    struct value *(*fun)();
    struct lnode *expr1;
    struct lnode *expr2;
};

#define	L_3			0x70000
struct lnode_3 {
    int line;
    struct value *(*fun)();
    struct lnode *expr1;
    struct lnode *expr2;
    struct lnode *expr3;
};

#define L_DEF			0x80000
#define L_STATIC		0x001
#define L_PRIVATE		0x002

struct lnode_def {
    int line;
    struct value *(*fun)();
    char *name;
    struct lnode *block;
    struct lnode_def *next;
    unsigned short num_var;	/* Number of local variables + arguments */
    unsigned char num_arg;	/* Number of arguments. */
    unsigned short num_ref;
    unsigned char flags;   
    int length;			/* String length of name. */
};

#define L_VAR_DEF		0x90000
struct lnode_var_def {
    int line;
    struct value *(*fun)();
    char *name;
    struct lnode_var_def *next;
    unsigned short num_var;
    unsigned char flags;
};

#define L_FUNCALL		0xa0000
struct lnode_funcall {
    int line;
    struct value *(*fun)();
    struct lnode *expr;
    char *name;
    struct lnode_funcall *next;
    struct fun_p *fun_p;
};

#define L_BLOCK			0xb0000
struct lnode_block {
    int line;
    struct value *(*fun)();
    int num_nodes;		/* Number of nodes in this block */
    char *block;
};

#define L_CONSTANT		0xc0000	/* Used for non-allocateable lnodes */

#define L_MASK			0xf0000
#define L_MAX			0xc0000	/* Highest number */
#define L_SHIFT			16	/* Number of bits to shift */
struct lnode {
    int line;
    struct value *(*fun)();
    struct lnode *a1, *a2, *a3;	/* Statements inside this statement */
};

struct inherit_list {
		char *name;
		struct lnode_def *prog;
		unsigned short num_var;
		int swap_num;
		struct lnode_var_def *status;
		struct lnode *heart_beat,*catch_tell;
		char *block;
		unsigned int swap_size;
	    struct inherit_list *next;
		int var_x_ref[1];
};

struct lnode_buff {
		char *fun;
		struct lnode_def *prog;
		struct inherit_list *list;
};

struct lnode_var_def *find_status PROT((char *));
struct inherit_list *alloc_inherit PROT((int));
void   free_inherit PROT((struct inherit_list *));
void   free_inherit_list PROT((struct inherit_list *,int));
struct lnode_number *alloc_lnode_number PROT((struct value *(*)(), int, double));
struct lnode_block *alloc_lnode_block PROT((struct lnode *));
struct lnode_name *alloc_lnode_name PROT((struct value *(*)(), char *));
struct lnode_variable *alloc_lnode_variable PROT((struct value *(*)(), int));
struct lnode_1 *alloc_lnode_1 PROT((struct value *(*)(), struct lnode *));
struct lnode_2 *alloc_lnode_2 PROT((struct value *(*)(), struct lnode *, struct lnode *));
struct lnode_3 *alloc_lnode_3 PROT((struct value *(*)(), struct lnode *, struct lnode *,
				    struct lnode *));
struct lnode_def *alloc_lnode_def PROT((struct value *(*)(), char *, struct lnode *,int));
struct lnode_funcall *alloc_lnode_funcall PROT((struct value *(*)(), char *, struct lnode *));
struct lnode_single *alloc_lnode_single PROT((struct value *(*)()));

extern struct lnode_var_def *prog_status;
int alloc_lnode_var_def PROT((struct value *(*)(), char *, int,int));
int free_prog PROT((struct lnode_def *, char *, int));
void free_sub_part PROT((struct lnode *, int)),
    add_prog_ref PROT((struct lnode_def *)),
    free_lnode PROT((struct lnode *, int));

extern int lnode_size[];
